import re


def is_paywalled(*, text: str) -> bool:
    """
    attempts to detect paywall by looking for certain strings
    in the text body of the article
    :param text: text body of article
    :return: True if paywall detected, False if unknown
    """
    if "are you a robot" in text.lower():
        return True
    if "please verify you're not a robot" in text.lower():
        return True
    if "what is included in my trial?" in text.lower():
        return True
    if re.search(r"You have (\d+(\.\d{1,2})?)% of this article left to read\. The rest is for subscribers only\.", text):
        return True
    return False





"""
paywall exceptions
case 1: article only half parses due to paywall (e.g. for NYT)
case 2: article doesn't parse at all due to paywall and has "are you a robot msg" or something similar (e.g. for Bloomberg, ft)
case 3: article doesn't parse at all due to paywall and no message (not where this is)
"""


class PaywallException(Exception):
    def __init__(self, message: str = "Partial paywall detected"):
        self.message = message
        super().__init__(self.message)


class BadURLException(Exception):
    def __init__(self, *, url: str):
        self.message = f"Bad url detected: {url}"
        super().__init__(self.message)





# -*- coding: utf-8 -*-
"""
Output formatting to text via lxml xpath nodes abstracted in this file.
"""
__title__ = 'np4k'
__author__ = 'Airvue'
__license__ = 'MIT'
__copyright__ = 'Copyright 2023, Airvue'

from html import unescape
import logging
from typing import Tuple, Union, Any
from urllib.parse import urlparse, urljoin

from lxml.html import HtmlElement
from markdownify import markdownify

from .text import innerTrim


log = logging.getLogger(__name__)


class OutputFormatter(object):

    def __init__(self, config):
        self.top_node = None
        self.config = config
        self.parser = self.config.get_parser()
        self.language = config.language
        self.stopwords_class = config.stopwords_class

    def update_language(self, meta_lang):
        '''Required to be called before the extraction process in some
        cases because the stopwords_class has to set incase the lang
        is not latin based
        '''
        if meta_lang:
            self.language = meta_lang
            self.stopwords_class = \
                self.config.get_stopwords_class(meta_lang)

    def get_top_node(self):
        return self.top_node

    def get_formatted(self, top_node: HtmlElement) -> tuple[str, Union[str, Any]]:
        """Returns the body text of an article, and also the body article
        html if specified. Returns in (text, html) form
        """
        # we might be able to remove this whole thing if we completely move to MD

        self.top_node = top_node
        html, text = '', ''

        self.remove_negativescores_nodes()

        if self.config.keep_article_html:
            html = self.convert_to_html()

        self.links_to_text()
        self.add_newline_to_br()
        self.add_newline_to_li()
        self.replace_with_text()
        self.remove_empty_tags()
        self.remove_trailing_media_div()
        text = self.convert_to_text()
        # print(self.parser.nodeToString(self.get_top_node()))
        return text, html

    def get_markdown(self, *, top_node: HtmlElement, source_url: str) -> str:
        """
        Returns the body text of an article in Markdown format
        design flaw: this can't be executed after get_body because it removes all the links
        """
        self.top_node = top_node
        self.convert_relative_links_to_absolute(base_url=source_url)
        md = markdownify(self.parser.nodeToString(top_node))
        return md

    def convert_relative_links_to_absolute(self, *, base_url: str):
        """
        converts all relative linked (e.g. href="/foo") to absolute links (e.g. href="https://www.apnews.com/foo")
        """
        for link in self.top_node.xpath(".//a[@href]"):
            href = link.get('href')
            parsed_url = urlparse(href)
            if not parsed_url.netloc:
                link.set('href', urljoin(base_url, href))


    def convert_to_text(self):
        txts = []
        for node in list(self.get_top_node()):
            try:
                txt = self.parser.getText(node)
            except ValueError as err:  # lxml error
                log.info('%s ignoring lxml node error: %s', __title__, err)
                txt = None

            if txt:
                txt = unescape(txt)
                txt_lis = innerTrim(txt).split(r'\n')
                txt_lis = [n.strip(' ') for n in txt_lis]
                txts.extend(txt_lis)
        return '\n\n'.join(txts)

    def convert_to_html(self):
        cleaned_node = self.parser.clean_article_html(self.get_top_node())
        return self.parser.nodeToString(cleaned_node)

    def add_newline_to_br(self):
        for e in self.parser.getElementsByTag(self.top_node, tag='br'):
            e.text = r'\n'

    def add_newline_to_li(self):
        for e in self.parser.getElementsByTag(self.top_node, tag='ul'):
            li_list = self.parser.getElementsByTag(e, tag='li')
            for li in li_list[:-1]:
                li.text = self.parser.getText(li) + r'\n'
                for c in self.parser.getChildren(li):
                    self.parser.remove(c)

    def links_to_text(self):
        """Cleans up and converts any nodes that should be considered
        text into text.
        """
        self.parser.stripTags(self.get_top_node(), 'a')

    def remove_negativescores_nodes(self):
        """If there are elements inside our top node that have a
        negative gravity score, let's give em the boot.
        """
        gravity_items = self.parser.css_select(
            self.top_node, "*[gravityScore]")
        for item in gravity_items:
            score = self.parser.getAttribute(item, 'gravityScore')
            score = float(score) if score else 0
            if score < 1:
                item.getparent().remove(item)

    def replace_with_text(self):
        """
        Replace common tags with just text so we don't have any crazy
        formatting issues so replace <br>, <i>, <strong>, etc....
        With whatever text is inside them.
        code : http://lxml.de/api/lxml.etree-module.html#strip_tags
        """
        self.parser.stripTags(
            self.get_top_node(), 'b', 'strong', 'i', 'br', 'sup')

    def remove_empty_tags(self):
        """It's common in top_node to exit tags that are filled with data
        within properties but not within the tags themselves, delete them
        """
        all_nodes = self.parser.getElementsByTags(
            self.get_top_node(), ['*'])
        all_nodes.reverse()
        for el in all_nodes:
            tag = self.parser.getTag(el)
            text = self.parser.getText(el)
            if (tag != 'br' or text != '\\r') \
                    and not text \
                    and len(self.parser.getElementsByTag(
                        el, tag='object')) == 0 \
                    and len(self.parser.getElementsByTag(
                        el, tag='embed')) == 0:
                self.parser.remove(el)

    def remove_trailing_media_div(self):
        """Punish the *last top level* node in the top_node if it's
        DOM depth is too deep. Many media non-content links are
        eliminated: "related", "loading gallery", etc. It skips removal if
        last top level node's class is one of NON_MEDIA_CLASSES.
        """

        NON_MEDIA_CLASSES = ('zn-body__read-all', )

        def get_depth(node, depth=1):
            """Computes depth of an lxml element via BFS, this would be
            in parser if it were used anywhere else besides this method
            """
            children = self.parser.getChildren(node)
            if not children:
                return depth
            max_depth = 0
            for c in children:
                e_depth = get_depth(c, depth + 1)
                if e_depth > max_depth:
                    max_depth = e_depth
            return max_depth

        top_level_nodes = self.parser.getChildren(self.get_top_node())
        if len(top_level_nodes) < 3:
            return

        last_node = top_level_nodes[-1]

        last_node_class = self.parser.getAttribute(last_node, 'class')
        if last_node_class in NON_MEDIA_CLASSES:
            return

        if get_depth(last_node) >= 2:
            self.parser.remove(last_node)



# -*- coding: utf-8 -*-
"""
Holds misc. utility methods which prove to be
useful throughout this library.
"""
__title__ = 'np4k'
__author__ = 'Airvue'
__license__ = 'MIT'
__copyright__ = 'Copyright 2023, Airvue'

import codecs
import hashlib
import logging
import os
import pickle
import random
import re
import string
import sys
from pathlib import Path
import threading
import time
from typing import Optional, List, Callable, Any, Generator, Dict

from hashlib import sha1

from bs4 import BeautifulSoup

from . import settings

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


class FileHelper:
    """
    Provides methods for working with files.
    """
    @staticmethod
    def loadResourceFile(filename: str) -> str:
        """
        Loads the contents of a resource file.

        :param filename: The name or path of the resource file.
        :return: The contents of the file as a string.
        :raises IOError: If the file cannot be opened or read.
        """
        if not os.path.isabs(filename):
            dirpath = os.path.abspath(os.path.dirname(__file__))
            path = os.path.join(dirpath, 'resources', filename)
        else:
            path = filename
        try:
            with codecs.open(path, 'r', 'utf-8') as f:
                content = f.read()
            return content
        except IOError:
            raise IOError(f"Couldn't open file {path}")


class ParsingCandidate:

    def __init__(self, url: str, link_hash: str):
        """
        Initializes a ParsingCandidate object.

        :param url: The URL of the candidate.
        :param link_hash: The MD5 hash of the URL.
        """
        self.url = url
        self.link_hash = link_hash


class RawHelper:
    """
    Provides methods for working with raw HTML data.
    """
    @staticmethod
    def get_parsing_candidate(url: str, raw_html: str) -> Optional[ParsingCandidate]:
        """
        Gets a parsing candidate object from a URL and raw HTML data.

        :param url: The URL of the web page.
        :param raw_html: The raw HTML data.
        :return: A ParsingCandidate object.
        """
        if isinstance(raw_html, str):
            raw_html = raw_html.encode('utf-8', 'replace')
        link_hash = f"{hashlib.md5(raw_html).hexdigest()}.{time.time()}"
        return ParsingCandidate(url=url, link_hash=link_hash)


class URLHelper:
    @staticmethod
    def get_parsing_candidate(url_to_crawl: str) -> ParsingCandidate:
        """Returns a ParsingCandidate object for the given URL to crawl.

        :param url_to_crawl: The URL to crawl.
        :return: A ParsingCandidate object.
        """
        # Replace shebang in urls
        final_url = url_to_crawl.replace('#!', '?_escaped_fragment_=') \
            if '#!' in url_to_crawl else url_to_crawl
        link_hash = f"{hashlib.md5(final_url.encode()).hexdigest()}.{time.time()}"
        return ParsingCandidate(url=final_url, link_hash=link_hash)


class StringSplitter:
    def __init__(self, pattern: str) -> None:
        self.pattern = re.compile(pattern)

    def split(self, string: str) -> list[str]:
        if not string:
            return []
        return self.pattern.split(string)


class StringReplacement:
    def __init__(self, pattern: str, replace_with: str) -> None:
        self.pattern = pattern
        self.replaceWith = replace_with

    def replaceAll(self, string: str) -> str:
        """
        Replaces all occurrences of `pattern` in `string` with `replace_with`.

        :param string: The string to perform replacements on.
        :return: The modified string with all replacements made.
        """
        if not string:
            return ''
        return string.replace(self.pattern, self.replaceWith)


class ReplaceSequence:
    def __init__(self):
        self.replacements: List[StringReplacement] = []

    def create(self, firstPattern: str, replaceWith: Optional[str] = None) -> StringReplacement:
        result = StringReplacement(firstPattern, replaceWith or '')
        self.replacements.append(result)
        return self

    def append(self, pattern: str, replaceWith: Optional[str] = None) -> StringReplacement:
        return self.create(pattern, replaceWith)

    def replaceAll(self, string: str) -> str:
        if not string:
            return ''

        mutatedString = string
        for rp in self.replacements:
            mutatedString = rp.replaceAll(mutatedString)
        return mutatedString


class TimeoutError(Exception):
    pass


def timelimit(timeout):
    """Borrowed from web.py, rip Aaron Swartz"""
    def wrapper(func: Callable[..., Any]) -> Callable[..., Any]:
        def wrapped(*args: Any, **kwargs: Any) -> Any:
            class Dispatch(threading.Thread):
                def __init__(self):
                    threading.Thread.__init__(self)
                    self.result = None
                    self.error = None

                    self.setDaemon(True)
                    self.start()

                def run(self):
                    try:
                        self.result = function(*args, **kw)
                    except:
                        self.error = sys.exc_info()
            c = Dispatch()
            c.join(timeout)
            if c.isAlive():
                raise TimeoutError("Function %s timed out after %s seconds" % (function.__name__, timeout))
            if c.error:
                raise c.error[0](c.error[1])
            return c.result
        return wrapped
    return wrapper


def domain_to_filename(domain: str) -> str:
    """All '/' are turned into '-', no trailing. schema's
    are gone, only the raw domain + ".txt" remains
    """
    filename = domain.replace('/', '-')
    if filename.endswith('-'):
        filename = filename[:-1]
    filename = f"{filename}.txt"
    return filename


def filename_to_domain(filename: str) -> str:
    """
    [:-4] for the .txt at end
    The filename should have the format "example.com.txt".
    :param filename: The filename to convert.
    :return: The domain name.
    """
    return filename.replace('-', '/')[:-4]


def is_ascii(word: str) -> bool:
    """Returns True if a word is only ASCII characters"""
    return all(ord(c) < 128 for c in word)


def extract_meta_refresh(html: str) -> Optional[str]:
    """ Parses html for a tag like:
    <meta http-equiv="refresh" content="0;URL='http://sfbay.craigslist.org/eby/cto/5617800926.html'" />
    Example can be found at: https://www.google.com/url?rct=j&sa=t&url=http://sfbay.craigslist.org/eby/cto/
    5617800926.html&ct=ga&cd=CAAYATIaYTc4ZTgzYjAwOTAwY2M4Yjpjb206ZW46VVM&usg=AFQjCNF7zAl6JPuEsV4PbEzBomJTUpX4Lg
    """
    soup = BeautifulSoup(html, 'html.parser')
    element = soup.find('meta', attrs={'http-equiv': 'refresh'})
    if element:
        try:
            wait_part, url_part = element['content'].split(";")
        except ValueError:
            # In case there are not enough values to unpack
            # for instance: <meta http-equiv="refresh" content="600" />
            return None
        else:
            # Get rid of any " or ' inside the element
            # for instance:
            # <meta http-equiv="refresh" content="0;URL='http://sfbay.craigslist.org/eby/cto/5617800926.html'" />
            if url_part.lower().startswith("url="):
                return url_part[4:].replace('"', '').replace("'", '')
    return None


def to_valid_filename(s: str) -> str:
    """
    Converts an arbitrary string (in our case a domain name) into a valid file name for caching.

    :param s: The string to convert.
    :return: The converted string.
    """
    valid_chars = "-_.() %s%s" % (string.ascii_letters, string.digits)
    return ''.join(c for c in s if c in valid_chars)


def cache_disk(seconds: int = 86400 * 5, cache_folder: str = "/tmp") -> Callable:
    """
    Decorator for caching function results to disk for a specified number of seconds.

    :param seconds: The number of seconds to cache the result for. Defaults to 5 days.
    :param cache_folder: The path to the cache folder. Defaults to "/tmp".
    :return: A decorator function that caches the results of a function to disk.
    """
    def do_cache(function: Callable) -> Callable:
        def inner_function(*args, **kwargs):
            """
            Calculates a cache key based on the decorated function signature.

            The domain of the inputs (args[1]) is used to hash the cache key.

            :param args: The positional arguments passed to the function.
            :param kwargs: The keyword arguments passed to the function.
            :return: The cached result or the result of the decorated function.
            """
            key = sha1((str(args[1]) + str(kwargs)).encode('utf-8')).hexdigest()
            filepath = Path(cache_folder) / key

            # Verify that the cached object exists and is less than X seconds old.
            if filepath.exists():
                modified = filepath.stat().st_mtime
                age_seconds = time.time() - modified
                if age_seconds < seconds:
                    with open(filepath, "rb") as file:
                        return pickle.load(file)

            # call the decorated function...
            result = function(*args, **kwargs)
            # ... and save the cached object for next time
            with open(filepath, "wb") as file:
                pickle.dump(result, file)
            return result
        return inner_function
    return do_cache


def print_duration(method: Callable) -> Callable:
    """
    Decorator for printing out the runtime duration of a method in seconds.

    :param method: The method to time.
    :return: The timed method.
    """
    def timed(*args, **kw):
        start_time = time.perf_counter()
        result = method(*args, **kw)
        end_time = time.perf_counter()
        duration = end_time - start_time
        print(f"{method.__name__} took {duration:.2f} seconds")
        return result
    return timed


def chunks(l: List[Any], n: int) -> Generator[List[Any], None, None]:
    """
    Yield n successive chunks from l.

    :param l: The list to split into chunks.
    :param n: The number of chunks to yield.
    :return: A generator that yields the successive chunks of the input list.
    """
    newn = int(len(l) / n)
    for i in range(0, n - 1):
        yield l[i * newn:i * newn + newn]
    yield l[n * newn - newn:]


def purge(fn, pattern):
    """Delete files in a dir matching pattern
    """
    for f in os.listdir(fn):
        if re.search(pattern, f):
            os.remove(os.path.join(fn, f))


def clear_memo_cache(source):
    """Clears the memoization cache for this specific news domain
    """
    d_pth = os.path.join(settings.MEMO_DIR, domain_to_filename(source.domain))
    if os.path.exists(d_pth):
        os.remove(d_pth)
    else:
        print('memo file for', source.domain, 'has already been deleted!')


def memoize_articles(source, articles):
    """When we parse the <a> links in an <html> page, on the 2nd run
    and later, check the <a> links of previous runs. If they match,
    it means the link must not be an article, because article urls
    change as time passes. This method also uniquifies articles.
    """
    source_domain = source.domain
    config = source.config

    if len(articles) == 0:
        return []

    memo = {}
    cur_articles = {article.url: article for article in articles}
    d_pth = os.path.join(settings.MEMO_DIR, domain_to_filename(source_domain))

    if os.path.exists(d_pth):
        f = codecs.open(d_pth, 'r', 'utf8')
        urls = f.readlines()
        f.close()
        urls = [u.strip() for u in urls]

        memo = {url: True for url in urls}
        # prev_length = len(memo)
        for url, article in list(cur_articles.items()):
            if memo.get(url):
                del cur_articles[url]

        valid_urls = list(memo.keys()) + list(cur_articles.keys())

        memo_text = '\r\n'.join(
            [href.strip() for href in (valid_urls)])
    # Our first run with memoization, save every url as valid
    else:
        memo_text = '\r\n'.join(
            [href.strip() for href in list(cur_articles.keys())])

    # new_length = len(cur_articles)
    if len(memo) > config.MAX_FILE_MEMO:
        # We still keep current batch of articles though!
        log.critical('memo overflow, dumping')
        memo_text = ''

    # TODO if source: source.write_upload_times(prev_length, new_length)
    ff = codecs.open(d_pth, 'w', 'utf-8')
    ff.write(memo_text)
    ff.close()
    return list(cur_articles.values())


def get_useragent() -> str:
    """
    Returns the next user agent in a saved file, chosen randomly.

    :return: The selected user agent string.
    """
    with open(settings.USERAGENTS, 'r') as f:
        agents = f.readlines()
        selection = random.randint(0, len(agents) - 1)
        agent = agents[selection]
        return agent.strip()


def get_available_languages() -> List[str]:
    """
    Returns a list of available languages and their 2 character input codes.

    :return: A list of 2 character language codes.
    """
    stopword_files = os.listdir(os.path.join(settings.STOPWORDS_DIR))
    two_dig_codes = [f.split('-')[1].split('.')[0] for f in stopword_files]
    for d in two_dig_codes:
        assert len(d) == 2
    two_dig_codes.sort()
    return two_dig_codes


def print_available_languages() -> None:
    """
    Prints available languages with their full names
    """
    language_dict = {
        'ar': 'Arabic',
        'be': 'Belarusian',
        'bg': 'Bulgarian',
        'da': 'Danish',
        'de': 'German',
        'el': 'Greek',
        'en': 'English',
        'es': 'Spanish',
        'et': 'Estonian',
        'fa': 'Persian',
        'fi': 'Finnish',
        'fr': 'French',
        'he': 'Hebrew',
        'hi': 'Hindi',
        'hr': 'Croatian',
        'hu': 'Hungarian',
        'id': 'Indonesian',
        'it': 'Italian',
        'ja': 'Japanese',
        'ko': 'Korean',
        'lt': 'Lithuanian',
        'mk': 'Macedonian',
        'nb': 'Norwegian (Bokmål)',
        'nl': 'Dutch',
        'no': 'Norwegian',
        'pl': 'Polish',
        'pt': 'Portuguese',
        'ro': 'Romanian',
        'ru': 'Russian',
        'sl': 'Slovenian',
        'sr': 'Serbian',
        'sv': 'Swedish',
        'sw': 'Swahili',
        'th': 'Thai',
        'tr': 'Turkish',
        'uk': 'Ukrainian',
        'vi': 'Vietnamese',
        'zh': 'Chinese',
    }

    codes = get_available_languages()
    print('\nYour available languages are:')
    print('\ninput code\t\tfull name')
    for code in codes:
        print(f'  {code}\t\t\t  {language_dict[code]}')
    print()


def extend_config(config: Any, config_items: Dict[str, Any]) -> Any:
    """
    Dynamically generates a new config object by extending an existing config object
    with the given key-value pairs.

    :param config: The existing config object to extend.
    :param config_items: The key-value pairs to add to the config object.
    :return: The new config object with the added key-value pairs.
    """
    for key, val in list(config_items.items()):
        if hasattr(config, key):
            setattr(config, key, val)

    return config



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['np4k', 'np4k.videos']

package_data = \
{'': ['*'], 'np4k': ['resources/misc/*', 'resources/text/*']}

install_requires = \
['beautifulsoup4>=4.11.2,<5.0.0',
 'cssselect>=1.2.0,<2.0.0',
 'feedfinder2>=0.0.4,<0.0.5',
 'feedparser>=6.0.10,<7.0.0',
 'jieba3k>=0.35.1,<0.36.0',
 'lxml>=4.9.2,<5.0.0',
 'markdownify>=0.11.6,<0.12.0',
 'nltk>=3.8.1,<4.0.0',
 'pillow>=9.4.0,<10.0.0',
 'pythainlp>=3.1.1,<4.0.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'pyyaml>=6.0,<7.0',
 'requests>=2.28.2,<3.0.0',
 'tinysegmenter>=0.4,<0.5',
 'tldextract>=3.4.0,<4.0.0']

setup_kwargs = {
    'name': 'np4k',
    'version': '0.0.95',
    'description': '',
    'long_description': '\n\nnp4k',
    'author': 'Airvue',
    'author_email': 'dev@airvue.news',
    'maintainer': 'Airvue',
    'maintainer_email': 'dev@airvue.news',
    'url': 'https://github.com/airvuetech/newspaper4k/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)

